import logging, malibu
from logging import handlers
from malibu.config import configuration
from malibu.util import get_caller

class LoggingDriver(object):

    CONFIG_SECTION = "logging"

    __instance = None

    @staticmethod
    def get_instance():

        if not LoggingDriver.__instance:
            return None

        return LoggingDriver.__instance

    @staticmethod
    def get_logger(name = None):

        if not LoggingDriver.get_instance():
            return None
        
        if name is None:
            name = get_caller()

        return LoggingDriver.get_instance().__get_logger(name = name)
    
    def __init__(self, config = {}, name = None):
        """ __init__(self, config = {}, name = None)

            Initializes the logging driver and loads necessary config
            values from the ConfigurationSection that should be passed
            in as config. Also loads the root logger from a specified
            default name or from the base module name of the package.
        """

        if not isinstance(config, configuration.ConfigurationSection):
            raise TypeError("Config should be of type "
                    "malibu.config.configuration.ConfigurationSection.")
        
        self.__config = config

        if not name:
            self.name = get_caller().split('.')[0]

        self.__logfile = self.__config.get_string("logfile", "/var/log/iocage-api.log")
        self.__loglevel = self.__config.get_string("loglevel", "INFO").upper()
        self.__stream = self.__config.get_bool("console_log", True)

        self.__loglevel = getattr(logging, self.__loglevel, None)
        if not isinstance(self.__loglevel, int):
            raise TypeError("Invalid log level: {}".format(
                self.__config.get_string("loglevel", "INFO").upper()))

        LoggingDriver.__instance = self

        self.__setup_logger()

    def __setup_logger(self):
        """ __setup_logger(self)

            Sets up the logging system with the logfile, loglevel, and
            other streaming options.
        """

        # Set up logging with the root handler so all log objects get the same
        # configuration.
        logger = logging.getLogger(self.name)
        logger.setLevel(self.__loglevel)
        formatter = logging.Formatter('%(asctime)s | %(name)-50s %(levelname)s : %(message)s')

        file_logger = handlers.RotatingFileHandler(
                self.__logfile,
                maxBytes = 8388608, # 8MB
                backupCount = 4)
        file_logger.setLevel(self.__loglevel)
        file_logger.setFormatter(formatter)

        logger.addHandler(file_logger)

        if self.__stream:
            logger.debug(" --> Building streaming log handler...")
            stream_logger = logging.StreamHandler()
            stream_logger.setLevel(self.__loglevel)
            stream_logger.setFormatter(formatter)

            logger.addHandler(stream_logger)

    def __get_logger(self, name = None):
        """ get_logger(self, name = None)

            Will return a logger object for a specific namespace. 
            If name parameter is None, get_logger will use call
            stack inspection to get the namespace of the last caller.
        """

        if name is None:
            name = get_caller()

        return logging.getLogger(name)



import struct, time, hmac

__packets__ = ['DataPacket']


def match_packet(data):

    get_class = lambda cls: globals()[cls]
    for packet in __packets__:
        try:
            packet = get_class(packet)
            p = packet()
            p.fromstr(data.split(DataPacket.__eop)[0])
            return p
        except (struct.error) as e: continue


def match_packets(data):

    for packet in data.split(DataPacket.__eop):
        yield match_packet(packet)


def register_packet(clsname):

    try: get_class(clsname)
    except: return

    __packets__.append(clsname)


class DataPacket(object):

    __eop = "\r\n"
    
    def __init__(self):
    
        self._format = ">i"
        
        self._pack = ['id']
    
        self.id = -1
        
    def fromstr(self, data):
    
        try:
            vals = self.unpack(data)
            ind = self._pack.index('id')
            if vals[ind] != self.id:
                raise struct.error("Packet id %s does not match my id %s" % (vals[ind], self.id))
            if len(vals) == len(self._pack):
                for var in self._pack:
                    setattr(self, var, vals[self._pack.index(var)])
                return vals
            else:
                raise struct.error("Data indexes do not match pack indexes.")
        except (struct.error) as e: raise

    def pack(self):
    
        vals = []
        for var in self._pack:
            vals.append(getattr(self, var))
        return struct.pack(self._format, *vals) + DataPacket.__eop
    
    def unpack(self, data):
    
        try: return struct.unpack(self._format, data)
        except (struct.error) as e: raise
        except: return None



import sqlite3, types, re, json, copy
from malibu.database import dbtypeconv


class DBMapper(object):

    # FETCH Constants for __execute()

    FETCH_ONE = 'one'
    FETCH_MANY = 'many'
    FETCH_ALL = 'all'
   
    # INDEX Constants for options dictionary.

    INDEX_PRIMARY = 'primaryIndex'
    INDEX_AUTOINCR = 'autoincrIndex'
    INDEX_UNIQUE = 'uniqueIndices'
    GENERATE_FTS_VT = 'genFTSVTs'

    # Global variables for static database methods

    _options = None
    __default_options = {
            INDEX_PRIMARY   : 0,
            INDEX_AUTOINCR  : True,
            INDEX_UNIQUE    : set(),
            GENERATE_FTS_VT : False # Do NOT generate FTS by default.
    }

    @staticmethod
    def get_default_options():

        return copy.deepcopy(DBMapper.__default_options)
    
    @staticmethod
    def connect_database(dbpath):

        dbtypeconv.install_json_converter()
        __db = sqlite3.connect(dbpath, detect_types = sqlite3.PARSE_DECLTYPES)

        return __db

    @classmethod
    def set_db_options(cls, db, keys, ktypes, options = __default_options):

        if cls._options is None:
            cls._options = {}
            cls._options['database'] = db
            cls._options['keys'] = keys
            cls._options['keytypes'] = ktypes
            cls._options['options'] = options
        else:
            cls._options['database'] = db
            cls._options['keys'] = keys
            cls._options['keytypes'] = ktypes

    @classmethod
    def load(cls, **kw):

        if cls._options is None:
            raise DBMapperException('Static database options have not been set.')

        dbo = cls._options
        obj = cls(dbo['database'])
        cur = dbo['database'].cursor()

        keys = []
        vals = []
        for key, val in kw.iteritems():
            keys.append(key)
            vals.append(val)
        whc = []
        for pair in zip(keys, vals):
            whc.append("%s=?" % (pair[0]))
        query = "select * from %s where (%s)" % (obj._table, ' and '.join(whc))
        result = obj.__execute(cur, query, args = vals)
        if result is None:
            for key in dbo['keys']:
                setattr(obj, "_%s" % (key), None)
            return
        for key, dbv in zip(dbo['keys'], result):
            setattr(obj, "_%s" % (key), dbv)

        return obj

    @classmethod
    def new(cls, **kw):

        if cls._options is None:
            raise DBMapperException('Static database options have not been set.')

        dbo = cls._options
        obj = cls(dbo['database'])
        cur = dbo['database'].cursor()

        keys = []
        vals = []
        for key, val in kw.iteritems():
            keys.append(key)
            vals.append(val)
        anonvals = []
        for val in vals:
            anonvals.append('?')
        query = "insert into %s (%s) values (%s)" % (obj._table, ','.join(keys), ','.join(anonvals))
        result = obj.__execute(cur, query, args = vals)
        
        return cls.load(**kw)

    @classmethod
    def find(cls, **kw):

        if cls._options is None:
            raise DBMapperException('Static database options have not been set.')
        
        dbo = cls._options
        obj = cls(dbo['database'])
        cur = dbo['database'].cursor()
        primaryKey = dbo['keys'][dbo['options'][DBMapper.INDEX_PRIMARY]]

        keys = []
        vals = []
        for key, val in kw.iteritems():
            keys.append(key)
            vals.append(val)
        whc = []
        for pair in zip(keys, vals):
            whc.append('%s=?' % (pair[0]))
        query = "select %s from %s where (%s)" % (primaryKey, obj._table, ' and '.join(whc))
        result = obj.__execute(cur, query, args = vals, fetch = DBMapper.FETCH_ALL)
        
        load_pairs = []
        for row in result:
            load_pairs.append({primaryKey : row[dbo['options'][DBMapper.INDEX_PRIMARY]]})

        return DBResultList([cls.load(**pair) for pair in load_pairs])

    @classmethod
    def find_all(cls):

        if cls._options is None:
            raise DBMapperException('Static database options have not been set.')
        
        dbo = cls._options
        obj = cls(dbo['database'])
        cur = dbo['database'].cursor()
        primaryKey = dbo['keys'][dbo['options'][DBMapper.INDEX_PRIMARY]]

        query = "select %s from %s" % (primaryKey, obj._table)
        result = obj.__execute(cur, query, fetch = DBMapper.FETCH_ALL)

        load_pairs = []
        for row in result:
            load_pairs.append({primaryKey : row[dbo['options'][DBMapper.INDEX_PRIMARY]]})

        return DBResultList([cls.load(**pair) for pair in load_pairs])

    @classmethod
    def search(cls, param):

        if cls._options is None:
            raise DBMapperException('Static database options have not been set.')

        if not cls._options['options'][DBMapper.GENERATE_FTS_VT]:
            raise DBMapperException('Full-text search table not enabled on this table.')

        dbo = cls._options
        obj = cls(dbo['database'])
        cur = dbo['database'].cursor()

        query = """select docid from _search_%s where _search_%s match \"?\"""" % \
                (obj._table, obj._table)

        result = obj.__execute(cur, query, args = [param], fetch = DBMapper.FETCH_ALL)

        load_pairs = []
        for row in result:
            load_pairs.append({cls._options['keys'][0] : row[0]})

        return DBResultList([cls.load(**pair) for pair in load_pairs])

    @classmethod
    def join(cls, cond, a, b):

        if cls._options is None or cond._options is None:
            raise DBMapperException('Static database options have not been set.')

        dba = cls._options
        obja = cls(dba['database'])
        dbb = cond._options
        objb = cond(dbb['database'])
        cur = dba['database'].cursor()
        primaryKeyA = dba['keys'][dba['options'][DBMapper.INDEX_PRIMARY]]
        primaryKeyB = dbb['keys'][dba['options'][DBMapper.INDEX_PRIMARY]]

        query = "select A.%s, B.%s from %s as A join %s as B on A.%s=B.%s" % (
                    primaryKeyA, primaryKeyB, obja._table, objb._table, a, b)
        result = obja.__execute(cur, query, fetch = DBMapper.FETCH_ALL)

        load_pair_a = []
        load_pair_b = []
        for row in result:
            load_pair_a.append({primaryKeyA : row[0]})
            load_pair_b.append({primaryKeyB : row[1]})

        return (DBResultList([cls.load(**pair) for pair in load_pair_a]), DBResultList([cond.load(**pair) for pair in load_pair_b]),)

    def __init__(self, db, keys, keytypes, options = __default_options):

        self._db = db
        self._options = options
        self._table = self.__class__.__name__.lower() if 'tableName' not in self._options else self._options['tableName']

        self._keys = keys
        self._keytypes = keytypes

        self._primary_ind = self._options[DBMapper.INDEX_PRIMARY]
        self._autoincr_ind = self._options[DBMapper.INDEX_AUTOINCR]
        self._primary = self._keys[self._primary_ind]
        self._unique_keys = self._options[DBMapper.INDEX_UNIQUE]

        self.__generate_structure()
        self.__generate_getters()
        self.__generate_setters()
        self.__generate_properties()

    def __execute(self, cur, sql, fetch = FETCH_ONE, limit = -1, args = ()):
        
        query = sql
        try:
            if len(args) >= 1:
                cur.execute("select " + ", ".join(["quote(?)" for i in args]), args)
                quoted_values = cur.fetchone()
                for quoted_value in quoted_values:
                    query = query.replace('?', str(quoted_value), 1)
        except: pass

        try: cur.execute(query)
        except (sqlite3.ProgrammingError):
            try: cur.execute(query, args)
            except Exception as e:
                raise DBMapperException("Error while executing query [%s]" % (query), cause = e)
        except Exception as e:
            raise DBMapperException("Error while executing query [%s]" % (query), cause = e)

        if fetch == DBMapper.FETCH_ONE:
            return cur.fetchone()
        elif fetch == DBMapper.FETCH_MANY:
            if limit == -1: limit = cur.arraysize
            return cur.fetchmany(size = limit)
        elif fetch == DBMapper.FETCH_ALL:
            return cur.fetchall()
        else:
            return cur.fetchall()

    def __get_table_info(self, table = None):

        table = self._table if table is None else table
        cur = self._db.cursor()
        query = "pragma table_info(%s)" % (table)

        return self.__execute(cur, query, fetch = DBMapper.FETCH_ALL)

    def __generate_structure(self):

        # use pragma constructs to get table into
        tblinfo = self.__get_table_info()

        # create the table if the statement does not exist
        if len(tblinfo) == 0:
            ins = zip(self._keys, self._keytypes)
            typarr = []
            for pair in ins:
                if pair[0] == self._primary:
                    # identifier type primary key
                    if self._autoincr_ind:
                        typarr.append("%s %s primary key autoincrement" % (pair[0], pair[1]))
                    else:
                        typarr.append("%s %s primary key" % (pair[0], pair[1]))
                elif pair[0] in self._unique_keys:
                    typarr.append("%s %s unique" % (pair[0], pair[1]))
                else:
                    # identifier type
                    typarr.append("%s %s" % (pair[0], pair[1]))
            cur = self._db.cursor()
            # create table if not exists <table> (<typarr>)
            query = "create table if not exists %s (%s)" % \
                (self._table, ', '.join(typarr))
            self.__execute(cur, query)

        # make sure table columns are up to date.
        if len(tblinfo) > 0:
            # use pragma table info to build database schema
            schema_ids = []
            schema_types = []
            for col in tblinfo:
                schema_ids.append(col[1])
                schema_types.append(col[2])
            # use schema to determine / apply database updates
            schema_updates = []
            for pair in zip(self._keys, self._keytypes):
                if pair[0] in schema_ids:
                    continue
                else:
                    schema_updates.append("%s %s" % (pair[0], pair[1]))
            for defn in schema_updates:
                query = "alter table %s add column %s" % (self._table, defn)
                cur = self._db.cursor()
                self.__execute(cur, query)

        # generate full text search table that corresponds with this dbo, if requested
        if self._options[DBMapper.GENERATE_FTS_VT]:
            if len(self.__get_table_info("_search_%s" % (self._table))) == 0:
                cur = self._db.cursor()
                # fts4 table doesn't exist, make it.
                query = "create virtual table _search_%s using fts4(%s, content='%s')" % \
                        (self._table, ','.join(self._keys), self._table)
                self.__execute(cur, query)
                # create pre/post update/delete triggers for cascading updates on content mod
                # XXX - [trigger warning] DO WE NEED THE TRIGGERS
                query = "create trigger _%s_bu before update on %s begin delete from _search_%s where docid=old.rowid; end;" % \
                        (self._table, self._table, self._table)
                self.__execute(cur, query)
                query = "create trigger _%s_bd before delete on %s begin delete from _search_%s where docid=old.rowid; end;" % \
                        (self._table, self._table, self._table)
                self.__execute(cur, query)
                search_keys = ','.join(['docid'] + self._keys[1:])
                target_keys = ','.join(['new.%s' % (vkey) for vkey in self._keys])
                query = "create trigger _%s_au after update on %s begin insert into _search_%s(%s) values(%s); end;" % \
                        (self._table, self._table, self._table, search_keys, target_keys)
                self.__execute(cur, query)
                query = "create trigger _%s_ai after insert on %s begin insert into _search_%s(%s) values(%s); end;" % \
                        (self._table, self._table, self._table, search_keys, target_keys)
                self.__execute(cur, query)

        self._db.commit()

    def __generate_getters(self):

        for _key in self._keys:
            def getter_templ(self, __key = _key):
                if __key not in self._keys:
                    return
                cur = self._db.cursor()
                # select * from table where key=<key>
                query = "select %s from %s where %s=?" % (__key, self._table, self._primary)
                result = self.__execute(cur, query, args = (getattr(self, "_%s" % (self._primary)),))
                try: return result[0]
                except: return result
            setattr(self, "get_%s" % (_key), types.MethodType(getter_templ, self))

    def __generate_setters(self):

        for _key in self._keys:
            def setter_templ(self, value, __key = _key):
                if __key not in self._keys:
                    return
                cur = self._db.cursor()
                # update table set key=value where primary=id
                query = "update %s set %s=? where %s=?" % (
                    self._table, __key, self._primary)
                self.__execute(cur, query, args = (value, getattr(self, "_%s" % (self._primary)),))
                setattr(self, "_%s" % (__key), value)
            setattr(self, "set_%s" % (_key), types.MethodType(setter_templ, self))

    def __generate_properties(self):

        for _key in self._keys:
            setattr(self, "_%s" % (_key), None)
        
        for _key in self._keys:
            getf = getattr(self, "get_%s" % (_key))
            setf = getattr(self, "set_%s" % (_key))
            setattr(self, _key, property(getf, setf, None, "[%s] property"))

    def create(self):

        cur = self._db.cursor()
        vals = []
        for key in self._keys:
            if key == self._primary and self._autoincr_ind:
                vals.append(None) # Put None in for the index since it's going to be autoincr'd
            else:
                vals.append(getattr(self, "_%s" % (key)))
        qst = ', '.join(["?" for item in vals])
        query = "insert into %s values (%s)" % (self._table, qst)
        self.__execute(cur, query, args = vals)
        setattr(self, "_%s" % (self._primary), cur.lastrowid)

    def delete(self):

        cur = self._db.cursor()
       
        qst = "%s=?" % (self._keys[self._primary_ind])
        primary_val = getattr(self, "_%s" % (self._keys[self._primary_ind]))

        query = "delete from %s where (%s)" % (self._table, qst)
        self.__execute(cur, query, args = (primary_val,) )

        del self


class DBResultList(list):

    def __init__(self, extend = None):

        if isinstance(extend, list):
            for item in extend:
                if isinstance(item, DBMapper):
                    self.append(item)
                else: continue

    def filter_equals(self, key, val):
        """ filter_equals(key, val) ->
              filters database find result based on
              key-value equality.
        """

        res = DBResultList()

        for dbo in self:
            try:
                if getattr(dbo, "_%s" % (key)) == val:
                    res.append(dbo)
                else: continue
            except: continue

        return res

    def filter_iequals(self, key, val):
        """ filter_iequals(key, val) ->
              filters database find result based on
              case insensitive key-value equality.
              assumes that db attribute and val are strings.
        """

        res = DBResultList()

        for dbo in self:
            try:
                if getattr(dbo, "_%s" % (key)).lower() == val.lower():
                    res.append(dbo)
                else: continue
            except: continue

        return res

    def filter_inequals(self, key, val):
        """ filter_inequals(key, val) ->
              filters database find result based on
              key-value inequality.
        """

        res = DBResultList()

        for dbo in self:
            try:
                if getattr(dbo, "_%s" % (key)) != val:
                    res.append(dbo)
                else: continue
            except: continue

        return res

    def filter_regex(self, key, regex):
        """ filter_regex(key, regex) ->
              filters database find result based on
              regex value matching.
        """

        res = DBResultList()

        for dbo in self:
            try:
                if re.match(regex, getattr(dbo, "_%s" % (key))) != None:
                    res.append(dbo)
                else: continue
            except: continue

        return res


class DBMapperException(Exception):

    def __init__(self, message, cause = None):

        if cause is not None:
            super(DBMapperException, self).__init__(message + u', caused by ' + repr(cause))
        elif cause is None:
            super(DBMapperException, self).__init__(message)

        self.message = message
        self.cause = cause

    def __str__(self):

        return repr(self.message) + u', caused by ' + repr(self.cause)



import socket, struct, ssl, time, collections
from packets import *


class TCPConnection(object):

    __eof = "\r\n"
    __connections = {}
    
    @staticmethod
    def get_connection(address):
    
        try: return TCPConnection.__connections[address]
        except: return None
    
    @staticmethod
    def get_connections():
    
        return TCPConnection.__connections.values()
    
    @staticmethod
    def remove_connection(address):

        try: return TCPConnection.__connections.pop(address)
        except: return None

    def __init__(self, (sock, address)):
    
        self._socket = sock
        self._address = address
        
        self._queue = collections.deque()
        
        self._identified = False
        
        TCPConnection.__connections.update({ self.get_address() : self })

    def fileno(self):
    
        return self._socket.fileno()

    def get_address(self):
    
        return "%s:%s" % (self._address)

    def send_next(self):
    
        try:
            res = [self._socket.send(data) for data in [self._queue.popleft() for count in xrange(0, 1)]]
            if 0 in res:
                TCPConnection.remove_connection(self.get_address())
        except (IndexError) as e: pass
        except: raise

    def send(self, data, now = False):
    
        data = self.connection_send_pre(data)

        if now:
            self._socket.send(data)
        else:
            self._queue.append(data)

        self.connection_send_post(data)
    
    def read_next(self):
    
        data = self._socket.recv(4096).split(TCPConnection.__eof)
        data.pop(len(data) - 1) # pop empty garbage data left over from splitting __eof
        for raw in data:
            packet = packets.match_packet(raw)
            if packet: self.connection_recv_packet(packet)
            else: self.connection_recv_data(raw)

    def connection_recv_packet(self, packet):

        pass

    def connection_recv_data(self, data):

        pass

    def connection_send_pre(self, data):

        pass

    def connection_send_post(self, data):

        pass



import glob, inspect, os

modules = glob.glob(os.path.dirname(__file__) + "/*.py")
__all__ = [os.path.basename(f)[:-3] for f in modules if not os.path.basename(f).startswith('_') and not f.endswith('__init__.py') and os.path.isfile(f)]


def get_caller():

    frame = inspect.currentframe()
    callstack = inspect.getouterframes(frame, 2)
    caller = callstack[2][0]
    callerinfo = inspect.getframeinfo(caller)
    
    if 'self' in caller.f_locals:
        caller_class = caller.f_locals['self'].__class__.__name__
    else:
        caller_class = None
    
    caller_module = inspect.getmodule(caller).__name__
    caller_name = callerinfo[2]
    
    if caller_class:
        caller_string = "%s.%s" % (caller_class, caller_name)
    else:
        caller_string = "%s" % (caller_name)

    if caller_module:
        caller_string = "%s." % (caller_module) + caller_string

    return caller_string



import datetime, traceback
from datetime import datetime, timedelta


class Scheduler(object):
    __instance = None
    
    @staticmethod
    def get_instance():
    
        if Scheduler.__instance is None:
            return Scheduler()
        else:
            return Scheduler.__instance

    def __init__(self):
    
        Scheduler.__instance = self
        
        self.__jobs = {}

    def create_job(self, name, function, delta, recurring = False):
    
        if name in self.__jobs:
            raise SchedulerException("Job already exists; remove it first.")
        if function is None:
            raise SchedulerException("Callback function is non-existent.")
        if not isinstance(delta, timedelta):
            raise SchedulerException("Argument 'delta' was not a timedelta instance.")
        
        job = SchedulerJob(name, function, delta, recurring)
        self.add_job(job)
        
        return job
    
    def add_job(self, job):
    
        if job in self.__jobs:
            raise SchedulerException("Job already exists; remove it first.")
        job.begin_ticking()
        self.__jobs.update({ job.get_name() : job })
    
    def remove_job(self, name):
    
        if not name in self.__jobs:
            raise SchedulerException("Job does not exist.")
        
        job = self.__jobs[name]
        
        self.__jobs.pop(name)
    
    def tick(self):
    
        now = datetime.now()
        
        for job in self.__jobs.values():
            if job.is_ready(now):
                try:
                    job.execute()
                    job.set_traceback(None)
                except:
                    job.set_traceback(traceback.extract_frame())
                if not job.is_recurring():
                    self.remove_job(job.get_name())
    

class SchedulerJob(object):

    def __init__(self, name, function, delta, recurring = False):

        self._scheduler = Scheduler.get_instance()
        self._name = name
        self._function = function
        self._delta = delta
        self._recurring = recurring
    	self._last_traceback = None
        
        self._eta = delta
    
    def get_name(self):
    
        return self._name
    
    def get_eta(self):
    
        return self._eta

    def get_traceback(self):

        return self._last_traceback

    def set_traceback(self, stack):

        self._last_traceback = stack
    
    def is_recurring(self):
    
        return self._recurring
    
    def is_ready(self, time):
    
        if time >= self._eta:
            return True
        else:
            return False
    
    def begin_ticking(self):

        self._eta = datetime.now() + self._delta
    
    def execute(self):
    
        self._function()
        
        if self._recurring:
            self._eta += self._delta


class SchedulerException(Exception):

    def __init__(self, value):

        self.value = value
    
    def __str__(self):

        return repr(self.value)



try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup

import malibu

setup(
    name = 'malibu',
    version = malibu.__version__,
    description = "maiome's library of utilities",

    url = "http://phabricator.maio.me/tag/malibu",
    author = "Sean Johnson",
    author_email = "sean.johnson@maio.me",
    
    license = "Unlicense",
    
    classifiers = [
        "Development Status :: 3 - Alpha",
        "License :: Public Domain",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules"
    ],
    packages = ['malibu',
                'malibu.config',
                'malibu.connection',
                'malibu.database',
                'malibu.text',
                'malibu.util'],
    package_dir = {'malibu': 'malibu'},
    zip_safe = True
)



import copy, urllib2, json
from contextlib import closing
from urllib2 import urlopen


""" We need to write some better documentation for this crap.
    Either that, or we just need to re-write the configuration stuff
    from scratch and actually document it this time.
"""


class ConfigurationSection(dict):
    """ modified dictionary that returns none for invalid keys
    """
    
    def __init__(self):

        dict.__init__(self)
        
        self.mutable = True
    
    def __getitem__(self, key):
    
        try: return dict.__getitem__(self, key)
        except (IndexError, KeyError) as e:
            raise KeyError("Unknown configuration key '%s'." % (key))
    
    def __setitem__(self, key, value):
    
        if self.mutable:
            self.update({ key: value })
        else:
            raise AttributeError("This section is not mutable.")
    
    def set_mutable(self, mutable):
    
        self.mutable = mutable
    
    def set(self, key, value):
    
        return self.__setitem__(key, value)
    
    def get(self, key):
    
        return self.__getitem__(key)
    
    def get_list(self, key, delimiter = ",", default = []):
    
        try:
            val = self.get(key)
            l = val.split(delimiter) if len(val) > 0 else default
            return [item.strip() for item in l]
        except: return default

    def get_string(self, key, default = ""):
    
        try:
            if str(self.get(key)) == '!None': return None
            return str(self.get(key)) or default
        except: return default
    
    def get_int(self, key, default = None):
    
        try: return int(self.get(key)) or default
        except: return default
    
    def get_bool(self, key, default = False):
    
        try:
            val = self.get(key) or default
            if isinstance(val, bool):
                return val
            elif isinstance(val, str):
                if val.lower() == 'true':
                    self.set(key, True)
                    return True
                elif val.lower() == 'false':
                    self.set(key, False)
                    return False
                else:
                    return default
            elif isinstanct(val, int):
                if val == 1:
                    self.set(key, True)
                    return True
                elif val == 0:
                    self.set(key, False)
                    return False
                else:
                    return default
            else:
                return default
        except: return default

class SectionPromise(object):
    """ this is a configuration section promise
        to make resolution of linked sections post-load
        easier.
    """
    
    promises = []
    
    def __init__(self, config, section, key, link):
        
        self.config = config
        self.section = section
        self.key = key
        self.link = link
        self.__fulfilled = False

        SectionPromise.promises.append(self)

    def __str__(self):
        """ Convert directly to a string for recreating the link
            during config write.  Better for serialization.
        """

        return '@' + self.link

    def resolve(self):
    
        if self.__fulfilled:
            return
        
        section = self.config.get_section(self.section)
        link = self.config.get_section(self.link)
        target = section.get(self.key)

        if isinstance(target, list):
            target.remove(self)
            target.append(link)
            section.set(self.key, target)
        else:
            section.set(self.key, link)

        # Preserve the promise for writing back out.
        section.set("_%s_promise" % (self.key), self)
        self.__fulfilled = True

class Configuration(object):
    """ 
        This definitely needs to be documented.
    """

    def __init__(self):
        """ initialise the container
            store in key:value format withing the certain category
        """

        self.__container = ConfigurationSection()
        
        self._filename = None

        self.loaded = False
        
    def __resolve_links(self):
        """ resolves all linked references.
        """
        
        for promise in SectionPromise.promises:
            promise.resolve()
        
        SectionPromise.promises = []

    def add_section(self, section_name):
        """ adds a new configuration section to the main dictionary.
        """
        
        section = ConfigurationSection()
        self.__container.set(section_name, section)
        
        return section
    
    def remove_section(self, section_name):
        """ removes a section from the main dictionary.
        """
        
        del self.__container[section_name]
    
    def sections(self):
        """ returns a list of all sections in the configuration.
        """

        return self.__container.keys()

    def has_section(self, section_name):
        """ return if x has a section
        """

        return section_name in self.__container

    def get_section(self, section_name):
        """ return a raw/direct reference to a section
        """

        return self.__container[section_name] if self.__container.__contains__(section_name) else None
    
    def unload(self):
        """ unload an entire configuration
        """

        self.__container.clear()
        self.loaded = False
    
    def reload(self):
        """ reload the configuration from the initially specified file
        """

        self.unload()
        self.load(self._filename)
    
    def save(self, filename = None):

        if filename is None:
            filename = self._filename

        if filename is None:
            raise ValueError('No filename specified and no stored filename.')

        with closing(open(filename, 'w')) as config:
            for section, smap in self.__container.iteritems():
                config.write("[%s]\n" % (section))
                for key, value in smap.iteritems():
                    if isinstance(value, list):
                        value = "+list:" + json.dumps(value)
                    elif isinstance(value, ConfigurationSection):
                        if "_%s_promise" % (key) in smap:
                            value = str(smap["_%s_promise" % (key)])
                        else:
                            value = str(value)
                    elif isinstance(value, SectionPromise):
                        continue
                    elif isinstance(value, file):
                        value = "+file:" + value.name
                    else:
                        value = str(value)
                    config.write("%s = %s\n" % (key, value))
                config.write("\n")

    def load(self, filename):
        """
            load(filename)
            
            filename -> name of the file to load.
        """

        try:
            fobj = open(filename, 'r')
            self._filename = filename
            self.load_file(fobj)
            fobj.close()
        except IOError as e:
            raise ValueError("Invalid filename '%s'." % (filename))
        except: 
            raise

    def load_file(self, fobj):
        """ load a file and read in the categories and variables
        """

        if self.loaded: self.__container.clear()
        section_name = None
        option_key = None
        option_value = None

        for line in fobj.readlines():
            line = line.strip('\n').lstrip()
        
            if line.startswith('#') or line.startswith('//') or line.startswith(';'):
                continue
            elif line.startswith('[') and line.endswith(']'):
                # This is the beginning of a section tag.
                section_name = line[1:-1]
                if not self.get_section(section_name):
                    self.add_section(section_name)
                continue
            elif '=' in line:
                set = line.split('=')
                l = len(set[0])
                # strip whitespace
                option_key = set[0].strip()
                option_value = set[1].lstrip() if set[1] is not '' or ' ' else None
                
                if option_value[-1] == ';': option_value = option_value[0:-1]
                section = self.get_section(section_name)
                
                if option_value.startswith('+'): # typed reference / variable
                    dobj_type = option_value.split(':')[0][1:]
                    if len(option_value.split(':')) > 2:
                        dobj_value = ':'.join(option_value.split(':')[1:])
                    else:
                        dobj_value = option_value.split(':')[1]
                    
                    if dobj_type.lower() == 'file':
                        try:
                            section.set(option_key, open(dobj_value, 'r'))
                        except:
                            try: section.set(option_key, open(dobj_value, 'w+'))
                            except: section.set(option_key, None)
                    elif dobj_type.lower() == 'url' or dobj_type.lower() == "uri":
                        try: section.set(option_key, urlopen(dobj_value).read())
                        except: 
                            raise
                            section.set(option_key, None)
                    elif dobj_type.lower() == 'list':
                        try: dobj_list = json.loads('%s' % (dobj_value))
                        except: dobj_list = []
                        dobj_repl = []
                        for item in dobj_list:
                            if item.startswith('@'):
                                link_name = item[1:]
                                if not self.get_section(link_name):
                                    dobj_repl.append(SectionPromise(self, section_name, option_key, link_name))
                                else:
                                    link = self.get_section(link_name)
                                    dobj_repl.append(link)
                            else: dobj_repl.append(item)
                        section.set(option_key, dobj_repl)
                    else:
                        section.set(option_key, option_value)
                elif option_value.startswith('@'): # section reference
                    link_name = option_value[1:]
                    if not self.get_section(link_name):
                        section.set(option_key, SectionPromise(self, section_name, option_key, link_name))
                    else:
                        link = self.get_section(link_name)
                        section.set(option_key, link)
                else: section.set(option_key, option_value)
                continue
            else:
                continue
        self.__resolve_links()
        self.loaded = True



import sys


class ArgumentParser(object):

    OPTION_SINGLE = 1
    OPTION_PARAMETERIZED = 2

    PARAM_SHORT = 1
    PARAM_LONG = 2

    @classmethod
    def from_argv(cls):
        """
            Creates an ArgumentParser engine with args from 
            sys.argv[].
        """
        return cls(sys.argv)

    def __init__(self, args, mapping = {}):
        """
            Initializes an ArgumentParser instance.  Parses out
            things that we already know, like args[0] (should be the 
            executable script).
        """

        try:
            self.exec_file = args[0]
        except:
            self.exec_file = ''

        self.__args = args
        
        self._default_types = {
                ArgumentParser.PARAM_SHORT : ArgumentParser.OPTION_SINGLE,
                ArgumentParser.PARAM_LONG  : ArgumentParser.OPTION_SINGLE
        }
        self._opt_types = {}
        self._mapping = mapping
        self._descriptions = {}
        self.options = {}
        self.parameters = []

    def set_default_param_type(self, param, opt = OPTION_SINGLE):
        """
            Sets the default type map that a parameter will be treated as.
            Can help force more uniform arguments without having to pre-define
            options.
        """

        self._default_types[param] = opt

    def add_option_type(self, option, opt = OPTION_SINGLE):
        """
            Adds a type mapping to a specific option. Allowed types are 
            OPTION_SINGLE and OPTION_PARAMETERIZED.
        """

        self._opt_types[option] = opt

    def add_option_mapping(self, option, map_name):
        """
            Maps a option value (-a, -g, --thing, etc.) to a full word or
            phrase that will be stored in the options dictionary after parsing
            is finished.  Makes option usage easier.
        """

        self._mapping[option] = map_name

    def add_option_description(self, option, description):
        """
            Adds a helpful description for an argument. Is returned when 
            get_option_descriptions is called.
        """

        self._descriptions[option] = description

    def get_option_descriptions(self):
        """
            Returns a map of options to their respective descriptions.  Good
            for printing out a series of help messages describing usage.
        """

        processed_descriptions = {}
        for option, description in self._descriptions.iteritems():
            if len(option) == 1: # This is a flag. Append a dash.
                option = '-' + option
                processed_descriptions.update({ option : description })
            elif len(option) > 1 and option.startswith('--'): # Option. Append blindly.
                processed_descriptions.update({ option : description })
            elif len(option) > 1 and not option.startswith('--'): # Option. Append double-dash.
                option = '--' + option
                processed_descriptions.update({ option : description })
            else: # What is this? Blindly append.
                processed_descriptions.update({ option : description })

        return processed_descriptions

    def parse(self):
        """ 
            Parses out the args into the mappings provided in self.mapping,
            stores the result in self.options.
        """

        waiting_argument = None

        for param in self.__args:
            if param.startswith('--'): # Long option mapping.
                paraml = param[2:]
                if '=' in paraml:
                    param, value = paraml.split('=')
                    # Automatically store a PARAMETERIZED option type, if none exists.
                    if param not in self._opt_types:
                        self.add_option_type(param, ArgumentParser.OPTION_PARAMETERIZED)
                else:
                    param = paraml
                    if param not in self._opt_types:
                        self.add_option_type(param, self._default_types[ArgumentParser.PARAM_LONG])
                store_as = param
                if param in self._mapping:
                    self.options.update({ self._mapping[param] : True })
                    store_as = self._mapping[param]
                else:
                    self.options.update({ param : True })
                if param in self._opt_types:
                    ptype = self._opt_types[param]
                    if ptype == ArgumentParser.OPTION_PARAMETERIZED:
                        if '=' in paraml: # Option is long and contains the value.
                            self.options.update({ store_as : value })
                            waiting_argument = None
                        else:
                            waiting_argument = store_as
            elif param.startswith('-'): # Short option mapping.
                param = param[1:]
                for opt in param:
                    store_as = opt
                    if opt in self._mapping:
                        self.options.update({ self._mapping[opt] : True })
                        store_as = self._mapping[opt]
                    else:
                        self.options.update({ opt : True })
                    if opt in self._opt_types:
                        ptype = self._opt_types[opt]
                        if ptype == ArgumentParser.OPTION_PARAMETERIZED:
                            waiting_argument = store_as
            else: # Option parameter or command parameter.
                if waiting_argument is not None:
                    self.options.update({ waiting_argument : param })
                    waiting_argument = None
                else:
                    self.parameters.append(param)



import socket, ssl, select, collections, time
from connection import TCPConnection


class TCPServer(object):

    __instance = None
    
    @staticmethod
    def get_instance():
    
        return TCPServer.__instance

    def __init__(self, host, port, ssl_enabled = False, ssl_keyfile = None, ssl_certfile = None):
        
        self._host = host
        self._port = port
        
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        
        self._shutdown = False
        
        if ssl_enabled:
            if not ssl_keyfile or not ssl_certfile:
                ssl_enabled = False
            else:
                self._socket = ssl.wrap_socket(self._socket, keyfile = self._ssl_keyfile, certfile = self._ssl_certfile)
        
        self._connections = []
        self._queue = collections.deque()
    
    def listen(self, backlog = 10):
    
        try:
            self._socket.bind((self._host, self._port))
            self._socket.listen(backlog)
        except (Exception) as e:
            self.server_uncaptured_exception(e)
    
    def run(self):
    
        while not self._shutdown:
            time.sleep(0.001)

            try:
                r, w, e = select.select([self._socket], [], [self._socket], 0.5)
                if self._socket in r:
                    client = TCPConnection(self._socket.accept())
                    self.client_connection_opened(client)

                if self._socket in e:
                    self.server_poll_error()
                    self._shutdown = True

                r, w, e = select.select(TCPConnection.get_connections(), [], [], 0.5)

                for connection in r:
                    try:
                        connection.read_next()
                        self.client_connection_read(self, connection)
                    except (socket.error) as e:
                        self.client_connection_lost(connection)
                        TCPConnection.remove_connection(connection.get_address())

                for connection in TCPConnection.get_connections():
                    connection.send_next()
            except (KeyboardInterrupt, SystemExit) as e:
                self._shutdown = True
            except (Exception) as e:
                self.server_uncaptured_exception(e)
        
        self._socket.shutdown(socket.SHUT_RDWR)
        self._socket.close()

    def client_connection_lost(self, connection):

        pass

    def client_connection_opened(self, connection):

        pass

    def client_connection_read(self, connection):

        pass

    def server_uncaptured_exception(self, exc):

        pass

    def server_poll_error(self):

        pass
